/*
    This is a QuickRes for Windows NT
    Copyright (C) 1995-2000  Andrey Shedel <andreys@cr.cyco.com>
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please send comments, corrections and contributions to andreys@cr.cyco.com

    The most recent version of this program is available from:
    http://www.chat.ru/~ashedel

*/



#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#ifndef UNICODE
#define UNICODE
#endif


#pragma warning(push, 3)
#include <windows.h>
#include <windowsx.h>
#include <shellapi.h>
#include <commctrl.h>
#pragma warning(pop)

#pragma warning(disable:4514) // '...' : unreferenced inline function has been removed

#pragma warning(push, 4)

#include "resource.h"
#include "msgfile.h"


#define IDM_FIRSTMODE 130
#define MAX_MODES     20


//
// 5 seconds to test new resolution.
//

#define TEST_TIMEOUT 5000


//
// Tray notify message
//

#define _WM_NOTIFYICON (WM_USER + 100)

static UINT WM_TASKBAR_CREATED;

//
// Icon id
//

#define ID_INDICATOR 100


//
// Misc. macros.
//


#ifndef DEFAULT_UNREACHABLE
#define DEFAULT_UNREACHABLE default: __assume(0)
#endif  DEFAULT_UNREACHABLE

#define ARGUMENT_PRESENT(v) (NULL != (v))

//
// misc. standard macroses.
//

#define FlagOn(Flags,SingleFlag)        ((SingleFlag) == ((Flags) & (SingleFlag)))

#define SetFlag(Flags,SingleFlag) { \
	(Flags) |= (SingleFlag);        \
}


#define ClearFlag(Flags,SingleFlag) { \
	(Flags) &= ~(SingleFlag);         \
}


#define NOTHING

#define NR_ELEMENTS(s) (sizeof(s)/sizeof(s[0]))




static const TCHAR BASEVIDEO[] = TEXT("VGA");


//
// Registry key/value strings.
//

#define SETTINGS_KEY_SOFTWARE  "Software"
#define COMPANY                "Cyco"
#define DEFINED_APPNAME        "TrayRes"


static const TCHAR APPNAME[]                 = TEXT(DEFINED_APPNAME);
static const TCHAR ERROR_READING_RESOURCES[] = TEXT("Error reading resources");
static const TCHAR MODETESTNAME[] = TEXT("ModeTestWindow");

static const TCHAR REGVAL_COUNT[] = TEXT("NumberOfItems");
static const TCHAR REGVAL_DEPTH[] = TEXT("Depth");
static const TCHAR REGVAL_X[] = TEXT("XResolution");
static const TCHAR REGVAL_Y[] = TEXT("YResolution");
static const TCHAR REGVAL_RATE[] = TEXT("RefreshRate");

static const TCHAR REGSTR_SETTINGS[] = TEXT(SETTINGS_KEY_SOFTWARE)TEXT("\\")TEXT(COMPANY)TEXT("\\")TEXT(DEFINED_APPNAME);

static const TCHAR GET_LAST_ERROR[] = TEXT("GetLastError");

static const TCHAR TB_CREATED_MESSAGE[] = TEXT("TaskbarCreated");


//
// Help file name
//

static const TCHAR HELPFILENAME[] = TEXT("qr.hlp>proc4");


//
// Tray window class.
//

static const TCHAR TRAYWINDOWCLASS[] = TEXT("Shell_TrayWnd");


//
//  Separator in listbox
//

#define ITEM_SEPARATOR 0xffffffffU

//
// structures
//

typedef struct __MODEPARAMS{
	UINT depth;
	UINT x;
	UINT y;
	UINT freq;
}MODEPARAMS, *PMODEPARAMS;

typedef struct __OPTIONSPAGEDATA{
	LPDEVMODE pvm;
	UINT c;
}OPTIONSPAGEDATA, *POPTIONSPAGEDATA;




//
// Globals
//


//HANDLE hinst; // resource module handle
#define GetResourceInstance() GetModuleHandle(NULL)
HWND   hwndOptions = NULL;



//
// Memory allocation.
// kernel32 heap used.
//


static
LPVOID
Realloc(
    IN LPVOID p,
    IN UINT cbBytes
  )
{
	return HeapReAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, p, cbBytes);
}


static
LPVOID
Alloc(
    IN UINT cbBytes
   )
{
	return HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, cbBytes);
}

static
VOID
Free(
    IN LPVOID p
   )
{
	HeapFree(GetProcessHeap(), 0, p);
}



//
// Debug helpers.
//

#if defined(NDEBUG)

#define DisplayError()
#define Trace

#else

static
void
DisplayError(void)
{
	DWORD error = GetLastError();

	if(ERROR_SUCCESS != error)
	{
		LPTSTR lpMsgBuf;

		FormatMessage(
					  FORMAT_MESSAGE_ALLOCATE_BUFFER |
					  FORMAT_MESSAGE_IGNORE_INSERTS |
					  FORMAT_MESSAGE_FROM_SYSTEM |
					  0,
					  NULL,
					  error,
					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
					  (LPTSTR)&lpMsgBuf, 0, NULL);

		if(NULL != lpMsgBuf)
		{
			MessageBox(NULL, lpMsgBuf, GET_LAST_ERROR, MB_OK|MB_ICONINFORMATION|MB_TASKMODAL);
			LocalFree((HLOCAL)lpMsgBuf);
		}
	}
}



static
void
__cdecl
Trace(
    IN LPTSTR lpszFormat,
    IN ... OPTIONAL
   )
{
	LPTSTR lpMsgBuf;
	va_list ap;

	va_start(ap, lpszFormat);


	FormatMessage(
				  FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_STRING,
				  lpszFormat, 0, 0, (LPTSTR)&lpMsgBuf, 0, &ap);

	if(NULL != lpMsgBuf)
	{
		OutputDebugString(lpMsgBuf);
		LocalFree((HLOCAL)lpMsgBuf);
	}
	else
	{
		DisplayError();
	}

	va_end(ap);

}

#endif





//
// Function used to check if new shell available.
//

static
BOOL
IsNewShell(VOID)
{
	//static int i = -1;
	int i;

	//if(i < 0)
	{
		OSVERSIONINFO vi = {sizeof(OSVERSIONINFO)};
		if(!GetVersionEx(&vi))
			return FALSE;

		i = (BOOL)(vi.dwMajorVersion >= 4);
	}

	return (BOOL)i;
}



//
// Check if tray window present.
//

static
BOOL
TrayPresent(VOID)
{
	return NULL != FindWindow(TRAYWINDOWCLASS, NULL);
}



//
// Check if I have to use tray window or minimized icon.
//

static
BOOL
AllowedToUseTray(VOID)
{
	return IsNewShell() && TrayPresent();
}




//
// Send message to the tray.
// If icon is passed it should be destroyed here.
//

static
BOOL
TrayMessage(
    IN HWND hDlg,
    IN DWORD dwMessage,
    IN UINT uID,
    IN HICON hIcon OPTIONAL,
    IN LPCTSTR  pszTip OPTIONAL
   )
{
	BOOL Succeeded;

	NOTIFYICONDATA tnd;

	tnd.cbSize     = sizeof(NOTIFYICONDATA);
	tnd.hWnd       = hDlg;
	tnd.uID        = uID;

	tnd.uCallbackMessage = _WM_NOTIFYICON;
	tnd.uFlags     = NIF_MESSAGE;

	if(ARGUMENT_PRESENT(hIcon))
	{
		SetFlag(tnd.uFlags, NIF_ICON);
		tnd.hIcon = hIcon;
	}

	if(ARGUMENT_PRESENT(pszTip))
	{
		SetFlag(tnd.uFlags, NIF_TIP);
		lstrcpyn(tnd.szTip, pszTip, NR_ELEMENTS(tnd.szTip));
	}

	Succeeded = Shell_NotifyIcon(dwMessage, &tnd);

	//
	// Destroy icon here
	//

	if (ARGUMENT_PRESENT(hIcon))
	{
		DestroyIcon(hIcon);
	}

	return Succeeded;
}




//
// Format message string.
//

static
LPCTSTR
__cdecl
FormatString(
    IN DWORD dwLang,
    IN HANDLE hMod,
    IN UINT id,
    OUT LPTSTR p,
    IN UINT cch,
    IN ... OPTIONAL
   )
{
	LPTSTR rv = p;
	va_list ap;
	va_start(ap, cch);

	if((0 == FormatMessage(
						   (p ? 0 : FORMAT_MESSAGE_ALLOCATE_BUFFER) |
						   FORMAT_MESSAGE_FROM_HMODULE, hMod, id,
						   dwLang,
						   (p ? p : (LPTSTR)&rv),
						   cch, &ap)) &&
	   (0 == FormatMessage(
						   (p ? 0 : FORMAT_MESSAGE_ALLOCATE_BUFFER) |
						   FORMAT_MESSAGE_FROM_HMODULE, hMod, id,
						   0,
						   (p ? p : (LPTSTR)&rv),
						   cch, &ap)))
	{
		DisplayError();

		if(NULL == p)
		{
			rv = LocalAlloc(LPTR, (lstrlen(ERROR_READING_RESOURCES) + 1) * sizeof(TCHAR));
			if(NULL != rv)
			{
				lstrcpy(rv, ERROR_READING_RESOURCES);
			}
		}
		else
		{
			lstrcpyn(rv, ERROR_READING_RESOURCES, cch);
		}
	}

	va_end(ap);

	return rv;
}


//
// Adjust the WS size.
//

static
void
AdjustWS(void)
{
	SetProcessWorkingSetSize(GetCurrentProcess(), (SIZE_T)-1, (SIZE_T)-1);
}




//
//	utility function. display message using string resources
//

static
UINT
__cdecl
StrResMessageBox(
    IN DWORD dwLang,
    IN HANDLE hMod,
    IN HWND hwnd,
    IN UINT uiText,
    IN UINT uiTitle,
    IN UINT fuiStyle,
    IN ... OPTIONAL
   )
{
	UINT rv;
	LPTSTR pszMsg = NULL;
	LPTSTR pszCaption = NULL;
	va_list ap;
	va_start(ap, fuiStyle);

	if(0 == FormatMessage(
						  FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE,
						  hMod, uiTitle,
						  dwLang,
						  (LPTSTR)&pszCaption, 0, &ap))
		FormatMessage(
					  FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE,
					  hMod, uiTitle,
					  0,
					  (LPTSTR)&pszCaption, 0, &ap);

	if(0 == FormatMessage(
						  FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE,
						  hMod, uiText,
						  dwLang,
						  (LPTSTR)&pszMsg, 0, &ap))
		FormatMessage(
					  FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE,
					  hMod, uiText,
					  0,
					  (LPTSTR)&pszMsg, 0, &ap);


	if((NULL != pszCaption) &&
	   (NULL != pszMsg))
	{
		rv = MessageBox(hwnd, pszMsg, pszCaption, fuiStyle);
	}
	else
	{
		DisplayError();
		rv = MessageBox(hwnd, ERROR_READING_RESOURCES, APPNAME, MB_OK | MB_ICONEXCLAMATION);
	}

	if(NULL != pszMsg)
	{
		LocalFree((HLOCAL)pszMsg);
	}

	if(NULL != pszCaption)
	{
		LocalFree((HLOCAL)pszCaption);
	}

	va_end(ap);

	return rv;
}





//
//	read/write registry values
//

//
//	write dword
//

static
VOID
WriteRegistryDword(
    IN HKEY hk,
    IN LPCTSTR szName,
    IN DWORD dwValue
   )
{
	RegSetValueEx(hk, szName, 0, REG_DWORD, (LPBYTE)&dwValue, sizeof(DWORD));
}


//
//	QueryRegistryDword
//

static
DWORD
QueryRegistryDword(
    IN HKEY hk,
    IN LPCTSTR szName,
    IN DWORD dwDefault
   )
{
	DWORD dwValueType;
	DWORD dwResult;
	DWORD dwDataSize = sizeof(DWORD);

	if((ERROR_SUCCESS ==
		RegQueryValueEx(hk,
						szName, 0, &dwValueType,
						(LPBYTE)&dwResult, &dwDataSize))
	   && (REG_DWORD == dwValueType))
	{
		return dwResult;
	}

	return dwDefault;
}





//
// Mode parameters helpers.
//


static
VOID
CopyModeParam(
    OUT PMODEPARAMS pModeParams,
    IN LPDEVMODE pvm
   )
{
	pModeParams->depth = pvm->dmBitsPerPel;
	pModeParams->x     = pvm->dmPelsWidth;
	pModeParams->y     = pvm->dmPelsHeight;
	pModeParams->freq  = pvm->dmDisplayFrequency;
}


static
BOOL
CompareModeParam(
    IN PMODEPARAMS pmode,
    IN LPDEVMODE pvm
   )
{
	return (0 != pmode->depth) && (0 != pmode->x) && (0 != pmode->y) &&
		(pmode->depth == pvm->dmBitsPerPel) &&
		(pmode->x     == pvm->dmPelsWidth) &&
		(pmode->y     == pvm->dmPelsHeight) &&
		(pmode->freq  == pvm->dmDisplayFrequency);
}


//
// Append devmode to list.
//

static
PMODEPARAMS
AppendModeParam(
    IN PMODEPARAMS pModeParams OPTIONAL,
    IN UINT size,
    IN LPDEVMODE pvm
   )
{
	if(ARGUMENT_PRESENT(pModeParams))
	{
		pModeParams = Alloc(sizeof(MODEPARAMS));
	}
	else
	{
		pModeParams = Realloc(pModeParams, sizeof(MODEPARAMS) * (size + 1));
	}


	if(NULL != pModeParams)
	{
		CopyModeParam(pModeParams + size, pvm);
	}

	return pModeParams;
}


//
// Delete entry at index.
//

static
PMODEPARAMS
DeleteModeParam(
    IN PMODEPARAMS pModeParams OPTIONAL,
    IN UINT size,
    IN UINT index
   )
{
	if(ARGUMENT_PRESENT(pModeParams) && (0 < size) && (size > index))
	{
		if(size > (index + 1))
		{
			CopyMemory(pModeParams + index, pModeParams + index + 1, sizeof(MODEPARAMS) * (size - index - 1));
		}

		--size;

		if(size > 0)
		{
			pModeParams = Realloc(pModeParams, size * sizeof(MODEPARAMS));
		}
		else
		{
			Free(pModeParams);
			pModeParams = NULL;
		}
	}

	return pModeParams;

}





//
// Convert depth bits to nr of colors.
//

static
UINT
Depth2Colors(
    IN UINT depth
   )
{
	UINT rv = 0;

	while ( depth > 0 )
	{
		rv |= 1UL << --depth;
	}

	return rv;
}



//
// Format string representing device mode.
//

static
LPCTSTR
FormatModeString(
    IN DWORD dwLang,
    IN HANDLE hMod,
    OUT LPTSTR p,
    IN UINT cch,
    IN DWORD x,
    IN DWORD y,
    IN DWORD freq,
    IN DWORD depth
   )
{
	UINT numcolors = Depth2Colors(depth);

	return FormatString(dwLang, hMod,
						(numcolors > 0x0000ffffUL) ?
						((freq > 1) ? MSG_MENUITEM_TRUE : MSG_MENUITEM_TRUE_DEFAULT) :
						((freq > 1) ? MSG_MENUITEM : MSG_MENUITEM_DEFAULT),
						p, cch, x, y, freq, numcolors + 1);
}



//
// Initialize devmode with some defaults.
//

static
void
FillDevmode(
    IN LPDEVMODE pvm
   )
{
	pvm->dmSize = sizeof(DEVMODE);
	pvm->dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT | DM_DISPLAYFREQUENCY | DM_DISPLAYFLAGS;
}



//
// Get currently selected mode.
//

static
BOOL
GetCurrentMode(
    OUT PMODEPARAMS pvm
   )
{
	HDC hdc = GetDC(NULL);

	if(NULL == hdc)
		return FALSE;

	pvm->x = GetDeviceCaps(hdc, HORZRES);
	pvm->y = GetDeviceCaps(hdc, VERTRES);
	pvm->depth = GetDeviceCaps(hdc, BITSPIXEL);
	pvm->freq = GetDeviceCaps(hdc, VREFRESH);

	ReleaseDC(NULL, hdc);

	return TRUE;
}



//
// Update the tray/title with current mode.
//

static
void
UpdateTrayState(
    IN HWND hwnd
   )
{
	TCHAR szTip[100];
	MODEPARAMS vm;

	GetCurrentMode(&vm);

	FormatModeString(GetThreadLocale(), GetResourceInstance(),
					 szTip, sizeof(szTip), vm.x, vm.y,
					 vm.freq, vm.depth);

	// send our tips
	if(AllowedToUseTray())
	{
		TrayMessage(hwnd, NIM_MODIFY, ID_INDICATOR,
					LoadImage(GetResourceInstance(), MAKEINTRESOURCE(IDI_MAINICON), IMAGE_ICON,
							  GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), 0),
					szTip);
	}
	else
	{
		SetWindowText(hwnd, szTip);
	}
}







//
// Read and fill in modes.
// Returns nr of detected modes.
//

static
UINT
FillVideoModes(
    IN LPCTSTR name,
    OUT LPDEVMODE * ppvm
   )
{
	UINT cModes = 0;

	if(ARGUMENT_PRESENT(ppvm))
	{
		//
		// Allocate initial storage.
		//

		*ppvm = Alloc(sizeof(DEVMODE));

		//
		// Enumerate them.
		//

		while((NULL != *ppvm) &&
			  (FillDevmode((*ppvm) + cModes),
			   EnumDisplaySettings(name, cModes, (*ppvm) + cModes)))
		{
			FillDevmode((*ppvm) + cModes);
			++cModes;
			*ppvm = Realloc(*ppvm, sizeof(DEVMODE) * (cModes + 1));
		}


		//
		// clean extra allocated memory
		//

		if(NULL != *ppvm)
		{
			if(cModes > 0)
			{
				*ppvm = Realloc(*ppvm, sizeof(DEVMODE) * cModes);
			}
			else
			{
				Free(*ppvm);
				*ppvm = NULL;
			}
		}


		if(NULL == *ppvm)
		{
			cModes = 0;
		}
	}

	//
	// And return to our caller.
	//

	return cModes;
}




//
// Switch mode to selected one.
// Only detected available modes are allowed.
//

static
LONG
SwitchMode(
    IN PMODEPARAMS pmode,
    IN LPDEVMODE pvm,
    IN UINT cModes
   )
{
	UINT i;

	for (i = 0; i < cModes; i++)
	{
		if(CompareModeParam(pmode, pvm + i))
		{
			return ChangeDisplaySettings(pvm + i, CDS_UPDATEREGISTRY);
		}
	}

	return DISP_CHANGE_BADMODE;
}



//
//	read/write registry
//

static
LPTSTR
GetModeKey(
    IN UINT index
   )
{
	UINT len = lstrlen(REGSTR_SETTINGS);
	UINT size = len * sizeof(TCHAR);
	LPTSTR pKey = Alloc(size + (20 * sizeof(TCHAR)));
	static const TCHAR format[] = TEXT("%u");

	if(NULL != pKey)
	{
		CopyMemory(pKey, REGSTR_SETTINGS, size);
		*(pKey + len) = TEXT('\\');

		wsprintf(pKey + len + 1, format, index);
	}

	return pKey;
}


static
BOOL
ReadRegistry(
    IN UINT index,
    IN PMODEPARAMS pmode
   )
{
	LPTSTR pKey = GetModeKey(index);

	pmode->depth = pmode->x = pmode->y = pmode->freq = 0;

	if(NULL != pKey)
	{
		HKEY hk;

		if(ERROR_SUCCESS ==
		   RegOpenKeyEx(HKEY_CURRENT_USER, pKey, 0, KEY_READ, &hk))
		{
			pmode->depth = QueryRegistryDword(hk, REGVAL_DEPTH, 0);
			pmode->x     = QueryRegistryDword(hk, REGVAL_X, 0);
			pmode->y     = QueryRegistryDword(hk, REGVAL_Y, 0);
			pmode->freq  = QueryRegistryDword(hk, REGVAL_RATE, 0);

			RegCloseKey(hk);
		}

		Free(pKey);
	}


	return
		(0 != pmode->depth) &&
		(0 != pmode->x) &&
		(0 != pmode->y);
}


static
void
WriteRegistry(
    IN UINT index,
    IN PMODEPARAMS pmode
   )
{
	LPTSTR pKey = GetModeKey(index);

	if(NULL != pKey)
	{
		HKEY hk;
		DWORD dwDisposition;

		if(ERROR_SUCCESS == RegCreateKeyEx(HKEY_CURRENT_USER, pKey, 0,
										   NULL, 0, KEY_WRITE, NULL,
										   &hk, &dwDisposition))
		{
			WriteRegistryDword(hk, REGVAL_DEPTH, pmode->depth);
			WriteRegistryDword(hk, REGVAL_X, pmode->x);
			WriteRegistryDword(hk, REGVAL_Y, pmode->y);
			WriteRegistryDword(hk, REGVAL_RATE, pmode->freq);

			RegCloseKey(hk);
		}

		Free(pKey);
	}

}






//
// Display DISP_xxx message.
//

static
void
DisplayResult(
    IN HWND hwnd,
    IN INT result
   )
{
	//
	// Get relared message code.
	//

	UINT messagecode;

	switch(result)
	{
	case DISP_CHANGE_SUCCESSFUL:

		//
		// no message on success.
		//

		//messagecode = MSG_DISP_CHANGE_SUCCESSFUL;
		//break;
		return;

	case DISP_CHANGE_RESTART:
		messagecode = MSG_DISP_CHANGE_RESTART;
		break;

	case DISP_CHANGE_BADFLAGS:
		messagecode = MSG_DISP_CHANGE_BADFLAGS;
		break;

	case DISP_CHANGE_FAILED:
		messagecode = MSG_DISP_CHANGE_FAILED;
		break;

	case DISP_CHANGE_BADMODE:
		messagecode = MSG_DISP_CHANGE_BADMODE;
		break;

	case DISP_CHANGE_NOTUPDATED:
		messagecode = MSG_DISP_CHANGE_NOTUPDATED;
		break;

	default:
		messagecode = MSG_DISP_CHANGE_FAILED;
	}

	//
	// And invoke messagebox.
	//

	StrResMessageBox(GetThreadLocale(), GetResourceInstance(), hwnd, messagecode, MSG_DISP_CHANGE_RESULTS, MB_ICONINFORMATION | MB_OK);

}




//
// Switch to mode by index.
//

static
BOOL
SwitchToIndex(
    IN HWND hwnd,
    IN UINT index
   )
{
	LPDEVMODE pvm = NULL;
	UINT cModes = FillVideoModes(NULL, &pvm);

	if(NULL != pvm)
	{
		MODEPARAMS mode;

		if(ReadRegistry(index, &mode))
		{
			DisplayResult(hwnd, SwitchMode(&mode, pvm, cModes));
		}

		Free(pvm);
	}

	return TRUE;
}


//
//	does pvm exists in pmodes
//

static
BOOL
ExistInSettings(
    IN LPDEVMODE pvm,
    IN PMODEPARAMS pmodes,
    IN UINT cParams,
    OUT UINT* pindex OPTIONAL
   )
{
	UINT i;
	for(i = 0; i < cParams; i++)
	{
		if(CompareModeParam(pmodes + i, pvm))
		{
			//
			// Found one.
			//

			if(ARGUMENT_PRESENT(pindex))
			{
				*pindex = i;
			}

			return TRUE;
		}
	}

	//
	// Does not exist.
	//

	return FALSE;

}





//
// Set consistent enable state for proppage buttons.
//

static
void
ReenableButtons(
    IN HWND hwnd
   )
{
	HWND hwndSys = GetDlgItem(hwnd, IDC_SYSTEM_MODES_LIST);
	HWND hwndSel = GetDlgItem(hwnd, IDC_SELECTED_MODES_LIST);
	UINT csys = ListBox_GetCount(hwndSys);
	UINT csel = ListBox_GetCount(hwndSel);
	//UINT possys = ListBox_GetCurSel(hwndSys);
	UINT possel = ListBox_GetCurSel(hwndSel);


	Button_Enable(GetDlgItem(hwnd, IDB_UP), (csel > 0) && (possel > 0));
	Button_Enable(GetDlgItem(hwnd, IDB_DOWN), (csel > 0) && ((possel + 1) < csel));
	Button_Enable(GetDlgItem(hwnd, IDB_DEL_MODE), csel > 0);
	Button_Enable(GetDlgItem(hwnd, IDB_ADD_MODE), (csys > 0) && (csel < MAX_MODES));
	Button_Enable(GetDlgItem(hwnd, IDB_TEST_MODE), csys > 0);
	Button_Enable(GetDlgItem(hwnd, IDB_ADD_SEPARATOR), (csel < MAX_MODES));
}




//
// Reset the page.
//

void
ResetPage(
    IN HWND hwnd
   )
{
	POPTIONSPAGEDATA p = (POPTIONSPAGEDATA)GetWindowLong(hwnd, GWL_USERDATA);
	HWND hwndSys = GetDlgItem(hwnd, IDC_SYSTEM_MODES_LIST);
	HWND hwndSel = GetDlgItem(hwnd, IDC_SELECTED_MODES_LIST);

	if(p && p->c && p->pvm && hwndSys && hwndSel)
	{
		HKEY hk;
		UINT cmp = 0;
		PMODEPARAMS pmp;
		UINT i;

		if(ERROR_SUCCESS ==
		   RegOpenKeyEx(HKEY_CURRENT_USER, REGSTR_SETTINGS, 0, KEY_READ, &hk))
		{
			cmp = QueryRegistryDword(hk, REGVAL_COUNT, 0);
			RegCloseKey(hk);
		}


		if(cmp > 0)
		{
			pmp = Alloc(sizeof(MODEPARAMS) * cmp);

			if(NULL == pmp)
			{
				cmp = 0;
			}
		}

        //
		// read registry
		//

		for(i = 0; i < cmp; i++)
			ReadRegistry(i, pmp + i);

		//
		// Clear listboxes.
		//

		ListBox_ResetContent(hwndSys);
		ListBox_ResetContent(hwndSel);

		//
		// add separators
		//

		{
			TCHAR string[100];

			FormatString(GetThreadLocale(), GetResourceInstance(), MSG_SEPARATOR, string, sizeof(string));

			for(i = 0; i < cmp; i++)
			{
				UINT index;
				index = ListBox_AddString(hwndSel, string);
				ListBox_SetItemData(hwndSel, index, ITEM_SEPARATOR);
			}
		}

		//
		// Fill in modex listbox.
		//

		for(i = 0; i < p->c; i++)
		{
			TCHAR string[100];
			MODEPARAMS mp;
			UINT index;
			HWND hwndInsertTo;

			CopyModeParam(&mp, p->pvm + i);

			FormatModeString(GetThreadLocale(), GetResourceInstance(),
							 string, sizeof(string), mp.x, mp.y,
							 mp.freq, mp.depth);

			if(ExistInSettings(p->pvm + i, pmp, cmp, &index))
			{
				hwndInsertTo = hwndSel;
				ListBox_DeleteString(hwndInsertTo, index);
				ListBox_InsertString(hwndInsertTo, index, string);
			}
			else
			{
				hwndInsertTo = hwndSys;
				index = ListBox_AddString(hwndInsertTo, string);
			}

			ListBox_SetItemData(hwndInsertTo, index, i);
		}

		ListBox_SetCurSel(hwndSys, 0);
		ListBox_SetCurSel(hwndSel, 0);
	}

	//
	// Refresh buttons state.
	//

	ReenableButtons(hwnd);

	//
	// Return to our caller.
	//

	return;
}




//
// PSN_APPLY handler.
//

static
void
ApplyPage(
    IN HWND hwnd
   )
{
	POPTIONSPAGEDATA p = (POPTIONSPAGEDATA)GetWindowLong(hwnd, GWL_USERDATA);
	HWND hwndSel = GetDlgItem(hwnd, IDC_SELECTED_MODES_LIST);

	if(p && p->c && p->pvm && hwndSel)
	{
		UINT c = ListBox_GetCount(hwndSel);
		UINT i;

		HKEY hk;
		DWORD dwDisposition;

		if(ERROR_SUCCESS == RegCreateKeyEx(HKEY_CURRENT_USER, REGSTR_SETTINGS, 0,
										   NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE | KEY_READ, NULL,
										   &hk, &dwDisposition))
		{
			UINT cOld = QueryRegistryDword(hk, REGVAL_COUNT, 0);

			WriteRegistryDword(hk, REGVAL_COUNT, c);
			RegCloseKey(hk);

			//
			// delete old values
			//

			for(i = 0; i < cOld; i++)
			{
				LPTSTR pKey = GetModeKey(i);

				if(NULL != pKey)
				{
					RegDeleteKey(HKEY_CURRENT_USER, pKey);
					Free(pKey);
				}
			}

			//
			// And save from list.
			//

			for(i = 0; i < c; i++)
			{
				UINT index = ListBox_GetItemData(hwndSel, i);
				MODEPARAMS mp;

				if((ITEM_SEPARATOR == index) || (index < p->c))
				{
					if(ITEM_SEPARATOR != index)
					{
						CopyModeParam(&mp, p->pvm + index);

						WriteRegistry(i, &mp);
					}
				}
			}
		}
	}

	//
	// Opdate PS state and return to our caller.
	//

	PropSheet_CancelToClose(GetParent(hwnd));
}


//
// Add modeline.
//

static
BOOL
AddMode(
    IN HWND hwnd
   )
{
	HWND hwndLB = GetDlgItem(hwnd, IDC_SYSTEM_MODES_LIST);
	POPTIONSPAGEDATA p = (POPTIONSPAGEDATA)GetWindowLong(hwnd, GWL_USERDATA);

	if(IsWindowEnabled(GetDlgItem(hwnd, IDB_ADD_MODE)) &&
	   p && p->c && p->pvm && hwndLB)
	{
		UINT index;
		UINT i;

		index = ListBox_GetCurSel(hwndLB);
		i = ListBox_GetItemData(hwndLB, index);
		ListBox_DeleteString(hwndLB, index);

		{
			UINT count = ListBox_GetCount(hwndLB);
			if(index >= count)
				index = (count == 0) ? 0 : (count - 1);

			ListBox_SetCurSel(hwndLB, index);
		}



		hwndLB = GetDlgItem(hwnd, IDC_SELECTED_MODES_LIST);

		if(hwndLB && (i < p->c))
		{
			MODEPARAMS mp;
			TCHAR string[100];

			CopyModeParam(&mp, p->pvm + i);

			FormatModeString(GetThreadLocale(), GetResourceInstance(),
							 string, sizeof(string), mp.x, mp.y,
							 mp.freq, mp.depth);

			index = ListBox_InsertString(hwndLB, ListBox_GetCurSel(hwndLB), string);
			ListBox_SetItemData(hwndLB, index, i);
			ListBox_SetCurSel(hwndLB, index);
		}

	}


	return TRUE;
}



//
// Delete modeline.
//

static
BOOL
DelMode(
    IN HWND hwnd
   )
{
	HWND hwndLB = GetDlgItem(hwnd, IDC_SELECTED_MODES_LIST);

	if(hwndLB)
	{
		UINT index;
		UINT i;
		index = ListBox_GetCurSel(hwndLB);
		i = ListBox_GetItemData(hwndLB, index);
		ListBox_DeleteString(hwndLB, index);

		{
			UINT count = ListBox_GetCount(hwndLB);
			if(index >= count)
				index = (count == 0) ? 0 : (count - 1);

			ListBox_SetCurSel(hwndLB, index);
		}

		if(ITEM_SEPARATOR != i)
		{
			hwndLB = GetDlgItem(hwnd, IDC_SYSTEM_MODES_LIST);
			if(hwndLB)
			{
				POPTIONSPAGEDATA p = (POPTIONSPAGEDATA)GetWindowLong(hwnd, GWL_USERDATA);
				if(p && p->c && p->pvm && (i < p->c))
				{
					TCHAR string[100];
					MODEPARAMS mp;

					CopyModeParam(&mp, p->pvm + i);

					FormatModeString(GetThreadLocale(), GetResourceInstance(),
									 string, sizeof(string), mp.x, mp.y,
									 mp.freq, mp.depth);

					index = ListBox_AddString(hwndLB, string);
					ListBox_SetItemData(hwndLB, index, i);
					ListBox_SetCurSel(hwndLB, index);
				}
			}
		}
	}


	return TRUE;
}


//
// Add the separator line.
//

static
BOOL
AddSep(
    IN HWND hwnd
   )
{
	HWND hwndSel = GetDlgItem(hwnd, IDC_SELECTED_MODES_LIST);

	if(hwndSel)
	{
		TCHAR string[100];
		UINT index;

		FormatString(GetThreadLocale(), GetResourceInstance(), MSG_SEPARATOR, string, sizeof(string));
		index = ListBox_GetCurSel(hwndSel);
		index = ListBox_InsertString(hwndSel, index, string);
		ListBox_SetItemData(hwndSel, index, ITEM_SEPARATOR);
		ListBox_SetCurSel(hwndSel, index);
	}

	return TRUE;
}







//
//	draw test table
//

static
void
DrawTable(
    IN HDC hdc,
    IN UINT x,
    IN UINT y,
    IN UINT cx,
    IN UINT cy
   )
{
	UINT i;
	HPEN hpen = (HPEN)SelectObject(hdc, GetStockObject(WHITE_PEN));

	#define GRID_STEP 40

	for(i = x; i < cx; i += GRID_STEP)
	{
		MoveToEx(hdc, i, y, NULL);
		LineTo(hdc, i, cy);
	}

	for(i = y; i < cy; i += GRID_STEP)
	{
		MoveToEx(hdc, x, i, NULL);
		LineTo(hdc, cx, i);
	}

	MoveToEx(hdc, x, cy, NULL);
	LineTo(hdc, cx, cy);
	LineTo(hdc, cx, y);

	SelectObject(hdc, hpen);

}



//
// Thread procedure for mode test desktop.
//

static
ULONG
__stdcall
TestDeskThreadProc(
    IN LPDEVMODE pvm
   )
{
	DWORD rv = 1;
	HDESK hdeskold = GetThreadDesktop(GetCurrentThreadId());
	static const TCHAR desktop_name[] = TEXT("MODE_TEST");

	if(hdeskold)
	{
		HDESK hdesk = CreateDesktop(desktop_name, NULL, NULL/*pvm*/, DF_ALLOWOTHERACCOUNTHOOK,
									DESKTOP_CREATEMENU | DESKTOP_CREATEWINDOW | DESKTOP_SWITCHDESKTOP |
									DESKTOP_ENUMERATE |DESKTOP_HOOKCONTROL | DESKTOP_READOBJECTS |
									DESKTOP_WRITEOBJECTS,
									NULL);


		if(hdesk)
		{
			if(SwitchDesktop(hdesk) && SetThreadDesktop(hdesk) &&
			   (DISP_CHANGE_SUCCESSFUL ==
				ChangeDisplaySettings(pvm, CDS_FULLSCREEN)))
			{
				HWND hwnd = GetDesktopWindow();
				HDC hdc = GetDC(hwnd);
				if(hdc)
				{
					RECT rc;
					GetClientRect(hwnd, &rc);
					DrawTable(hdc, rc.left, rc.top, rc.right - 1, rc.bottom - 1);
					ReleaseDC(hwnd, hdc);
				}

				Sleep(TEST_TIMEOUT);
				rv = 0;
			}

			SwitchDesktop(hdeskold);
			SetThreadDesktop(hdeskold);

			CloseDesktop(hdesk);

		}
		else
		{
			DisplayError();
		}
	}

	ExitThread(rv);
	return 0;
}




//
// Test currently selected modeline.
//

static
BOOL
TestMode(
    IN HWND hwnd
   )
{
	HWND hwndLB = GetDlgItem(hwnd, IDC_SYSTEM_MODES_LIST);
	POPTIONSPAGEDATA p = (POPTIONSPAGEDATA)GetWindowLong(hwnd, GWL_USERDATA);
	INT result = DISP_CHANGE_BADMODE;

	if(p && p->c && p->pvm && hwndLB)
	{
		UINT index;
		UINT i;

		index = ListBox_GetCurSel(hwndLB);
		i = ListBox_GetItemData(hwndLB, index);

		if(i < p->c)
		{
			LPDEVMODE pvm = p->pvm + i;

			//
			// test if valid mode
			//

			result = ChangeDisplaySettings(pvm, CDS_TEST);

			if(DISP_CHANGE_SUCCESSFUL == result)
			{
				//
				// Test succeeded,
				// Do it...
				//

				DEVMODE vmcurrent;
				MODEPARAMS mp;

				//
				// But only if current is not the same.
				//

				if(!GetCurrentMode(&mp))
				{
					return FALSE;
				}

				CopyMemory(&vmcurrent, pvm, sizeof(DEVMODE));
				vmcurrent.dmPelsWidth = mp.x;
				vmcurrent.dmPelsHeight = mp.y;
				vmcurrent.dmBitsPerPel = mp.depth;
				vmcurrent.dmDisplayFrequency = mp.freq;
				vmcurrent.dmSize = sizeof(DEVMODE);
				vmcurrent.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT | DM_DISPLAYFREQUENCY;

				//
				// Perform visual test
				//

				if(IDOK != StrResMessageBox(GetThreadLocale(), GetResourceInstance(), hwnd,
											MSG_TESTING_MODE_TEXT,
											MSG_TESTING_MODE_TITLE,
											MB_ICONQUESTION | MB_OKCANCEL))
				{
					//
					// User cancelled the test.
					//

					return FALSE;
				}

				//
				// Finally do it.
				//

				{
					DWORD dwIdThread;
					HANDLE hthread = CreateThread(NULL, 8192,
												  TestDeskThreadProc,
												  pvm,
												  0,
												  &dwIdThread);

					if(NULL  != hthread)
					{
						//
						// Wait for it.
						//

						DWORD exitcode = 1;
						WaitForSingleObject(hthread, TEST_TIMEOUT * 5);

						if(GetExitCodeThread(hthread, &exitcode) &&
						   (0 == exitcode))
						{
							result = DISP_CHANGE_SUCCESSFUL;
						}

						CloseHandle(hthread);
					}

				}

			}
		}
	}

	//
	// Show some final message.
	//

	DisplayResult(hwnd, result);

	//
	// And return to our caller.
	//

	return DISP_CHANGE_SUCCESSFUL == result;
}



//
// Move the item in selected modes list.
//

static
void
MoveItem(
    IN HWND hwnd,
    IN BOOL up
   )
{
	if(hwnd)
	{
		UINT index = ListBox_GetCurSel(hwnd);
		TCHAR string[100];

		if(ListBox_GetTextLen(hwnd, index) < sizeof(string))
		{
			UINT data = ListBox_GetItemData(hwnd, index);
			ListBox_GetText(hwnd, index, string);
			ListBox_DeleteString(hwnd, index);

			if(up)
			{
				if(index > 0)
					--index;
			}
			else
			{
				UINT count = ListBox_GetCount(hwnd);
				if(index < count)
					++index;
			}

			ListBox_InsertString(hwnd, index, string);
			ListBox_SetItemData(hwnd, index, data);
			ListBox_SetCurSel(hwnd, index);
		}
	}
}


//
//	handle WM_COMMAND
//

static
BOOL
OptionsPageProc_OnCommand(
    IN HWND hwnd,
    IN int id,
    IN HWND hwndCtl,
    IN UINT codeNotify
   )
{
	UNREFERENCED_PARAMETER(hwndCtl);

	switch ( id )
	{
	case IDC_SYSTEM_MODES_LIST:
	case IDC_SELECTED_MODES_LIST:
		if((LBN_DBLCLK == codeNotify) &&
		   ((IDC_SYSTEM_MODES_LIST == id) ?
			AddMode(hwnd) : DelMode(hwnd)))
		{
			break;
		}

		if(LBN_SELCHANGE == codeNotify)
		{
			ReenableButtons(hwnd);
		}

		return FALSE;

	case IDB_UP:
	case IDB_DOWN:
		MoveItem(GetDlgItem(hwnd, IDC_SELECTED_MODES_LIST), IDB_UP == id);
		break;


	case IDB_TEST_MODE:
		TestMode(hwnd);
		return TRUE;

	case IDB_ADD_MODE:
		AddMode(hwnd);
		break;

	case IDB_DEL_MODE:
		DelMode(hwnd);
		break;

	case IDB_ADD_SEPARATOR:
		AddSep(hwnd);
		break;

	default:
		return FALSE;
	}

	//
	// Refresh buttons and set changed flag.
	//

	ReenableButtons(hwnd);
	PropSheet_Changed(GetParent(hwnd), hwnd);

	//
	// Return to our caller.
	//

	return TRUE;
}




//
//  Handle options page notifications
//

static
BOOL
OptionsPageProc_OnNotify(
    IN HWND hwnd,
    IN WPARAM wParam,
    IN NMHDR FAR* lpnmhdr
   )
{
	UNREFERENCED_PARAMETER(wParam);

	switch (lpnmhdr->code)
	{
	case PSN_KILLACTIVE:

		//
		// need PSN_APPLY
		//

		SetWindowLong(hwnd, DWL_MSGRESULT, (LONG)FALSE);
		break;

	case PSN_APPLY:
		ApplyPage(hwnd);
		SetWindowLong(hwnd, DWL_MSGRESULT, (LONG)PSNRET_NOERROR);
		break;

	case PSN_RESET:
		ResetPage(hwnd);
		break;

	case PSN_HELP:
		WinHelp(hwnd, HELPFILENAME, HELP_CONTEXT, IDD_SELECT_MODES);
		break;

	default:
		return FALSE;
	}

	//
	// Handled.
	//

	return TRUE;
}


//
// Options dialog procedure.
//

static
BOOL
CALLBACK
OptionsPageProc(
    IN HWND hwnd,
    IN UINT uMsg,
    IN WPARAM wParam,
    IN LPARAM lParam
   )
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		SetWindowLong(hwnd, GWL_USERDATA, (LONG)(((PROPSHEETPAGE *)lParam)->lParam));
		hwndOptions = GetParent(hwnd);
		ResetPage(hwnd);
		return TRUE;

	case WM_NOTIFY:
		return HANDLE_WM_NOTIFY(hwnd, wParam, lParam, OptionsPageProc_OnNotify);

	case WM_COMMAND:
		return HANDLE_WM_COMMAND(hwnd, wParam, lParam, OptionsPageProc_OnCommand);

	default:
		return FALSE;
	}

	return TRUE;
}







//
// Invoke options dialog.
//

static
void
ChangeOptions(
    void
   )
{
	if(!hwndOptions ||
	   !IsWindow(hwndOptions) ||
	   !SetForegroundWindow(hwndOptions))
	{

		static BOOL fInThere = FALSE;

		if(fInThere)
		{
			return;
		}

		fInThere = TRUE;

		{
			LPDEVMODE pvm = NULL;
			UINT cModes = FillVideoModes(NULL, &pvm);
			OPTIONSPAGEDATA p;

			PROPSHEETPAGE psp;
			PROPSHEETHEADER ph;

			psp.dwSize = PROPSHEETPAGEA_V1_SIZE;
			psp.dwFlags = PSP_DEFAULT | PSP_HASHELP;
			psp.hInstance = GetResourceInstance();
			psp.pszTemplate = MAKEINTRESOURCE(IDD_SELECT_MODES);
			psp.pszIcon = NULL;
			psp.pszTitle = NULL;
			psp.pfnDlgProc = OptionsPageProc;
			psp.lParam = (LPARAM)&p;
			psp.pfnCallback = NULL;
			psp.pcRefParent = NULL;

			ph.dwSize = PROPSHEETHEADER_V1_SIZE;
			ph.dwFlags = PSH_DEFAULT | PSH_USEICONID | PSH_PROPTITLE | PSH_PROPSHEETPAGE;
			ph.hwndParent = NULL;
			ph.hInstance = GetResourceInstance();
			ph.pszIcon = MAKEINTRESOURCE(IDI_MAINICON);
			ph.pszCaption = MAKEINTRESOURCE(IDS_APPNAME);
			ph.nPages = 1;
			ph.nStartPage = 0;
			ph.ppsp = &psp;
			ph.pfnCallback = NULL;

			p.c = cModes;
			p.pvm = pvm;

			(void)PropertySheet(&ph);

			if(NULL != pvm)
			{
				Free(pvm);
			}

		}

		fInThere = FALSE;
	}
}





//
// Show about dialog.
//

static
void
About(void)
{
	static BOOL fInThere = FALSE;

	if(!fInThere)
	{
		LPCTSTR psz;
		LPCTSTR pszTitle;

		fInThere = TRUE;

		psz = FormatString(GetThreadLocale(), GetResourceInstance(), MSG_ABOUTTEXT, NULL, 0);
		pszTitle = FormatString(GetThreadLocale(), GetResourceInstance(), MSG_APPNAME, NULL, 0);

		if((NULL != pszTitle) &&
		   (NULL != psz))
		{
			ShellAbout(NULL,
					   pszTitle, psz,
					   LoadIcon(GetResourceInstance(), MAKEINTRESOURCE(IDI_MAINICON)));
		}

		if(NULL != pszTitle)
			LocalFree((HLOCAL)pszTitle);

		if(NULL != psz)
			LocalFree((HLOCAL)psz);

		fInThere = FALSE;
	}
}



//
//	does mode exist in the list?
//

static
BOOL
ExistInList(
    IN PMODEPARAMS pmode,
    IN LPDEVMODE pvm,
    IN UINT cModes
   )
{
	UINT i;
	for(i = 0; i < cModes; i++)
	{
		if(CompareModeParam(pmode, pvm + i))
		{
			return TRUE;
		}
	}

	return FALSE;
}




//
//	fill popup menu
//

void
FillMenu(
    IN HMENU hMenu
   )
{
	LPDEVMODE pvm;
	UINT cModes = FillVideoModes(NULL, &pvm);

	if(NULL != pvm)
	{
		HKEY hk;

		if(ERROR_SUCCESS ==
		   RegOpenKeyEx(HKEY_CURRENT_USER, REGSTR_SETTINGS, 0, KEY_READ, &hk))
		{
			UINT count = QueryRegistryDword(hk, REGVAL_COUNT, 0);
			MODEPARAMS mode_current;

			GetCurrentMode(&mode_current);

			while(count > 0)
			{
				MODEPARAMS mode;

				--count;

				if(ReadRegistry(count, &mode))
				{
					LPCTSTR psz =
						FormatModeString(GetThreadLocale(), GetResourceInstance(), NULL, 0,
										 mode.x, mode.y, mode.freq, mode.depth);
					if(NULL != psz)
					{
						BOOL fCheck  = (mode.x == mode_current.x) &&
							(mode.y == mode_current.y) &&
							(mode.freq == mode_current.freq) &&
							(mode.depth == mode_current.depth);
						BOOL fEnable = !ExistInList(&mode, pvm, cModes);

						InsertMenu(hMenu, 0, MF_BYPOSITION | (fCheck ? MF_CHECKED : 0) |
								   (fEnable ? MF_GRAYED : 0), count + IDM_FIRSTMODE, psz);

						LocalFree((HLOCAL)psz);
					}
				}
				else
				{
					InsertMenu(hMenu, 0, MF_BYPOSITION | MF_SEPARATOR, 0, NULL);
				}

			}

			RegCloseKey(hk);

		}

		Free(pvm);
	}

}





//
// This one is useful when no tray is available.
//

static
void
ReplaceSystemMenu(
    IN HINSTANCE hinst,
    IN HWND hwnd,
    IN UINT id,
    IN UINT idDefault
   )
{
	HMENU hNewMenu = LoadMenu(hinst, MAKEINTRESOURCE(id));

	if(NULL != hNewMenu)
	{
		HMENU hNewSubMenu = GetSubMenu(hNewMenu, 0);

		if(NULL != hNewSubMenu)
		{
			HMENU hMenu = GetSystemMenu(hwnd, FALSE);

			if(AllowedToUseTray() && TrayPresent())
			{
				//
				// clear it first
				//

				while(GetMenuItemCount(hMenu) > 0)
					RemoveMenu(hMenu, 0, MF_BYPOSITION);

				{
					int i;
					int base = GetMenuItemCount(hMenu);

					if(base < 0)
						base = 0;

					for(i = 0; i < GetMenuItemCount(hNewSubMenu); i++)
					{
						MENUITEMINFO mii;
						TCHAR buf[1024];

						mii.cbSize = sizeof(mii);
						mii.dwTypeData = &(buf[0]);
						mii.cch = sizeof(buf);
						mii.fMask = MIIM_DATA | MIIM_ID | MIIM_STATE | MIIM_SUBMENU |
							MIIM_TYPE | MIIM_CHECKMARKS;

						if(GetMenuItemInfo(hNewSubMenu, i, TRUE, &mii))
							InsertMenuItem(hMenu, i + base, TRUE, &mii);
					}

				}

				if(0 != idDefault)
				{
					SetMenuDefaultItem(hMenu, idDefault, FALSE);
				}

			}
			else
			{
				TCHAR szPopupName[100];

				GetMenuString(hNewMenu, 0, szPopupName, sizeof(szPopupName), MF_BYPOSITION);
				AppendMenu(hMenu, MF_SEPARATOR, 0, NULL);
				AppendMenu(hMenu, MF_POPUP, (UINT)hNewSubMenu, szPopupName);
			}

			FillMenu(hNewSubMenu);

		}

		DestroyMenu(hNewMenu);
	}

}




static
VOID
ProcessMainWindowMenu(
    IN HWND hwnd
   )
{
	HMENU hPopupMenu;
	HMENU hMenu;

	hMenu = LoadMenu(GetResourceInstance(), MAKEINTRESOURCE(IDR_POPMENU));

	if(NULL == hMenu)
	{
		DestroyWindow(hwnd);
	}

	hPopupMenu = GetSubMenu(hMenu, 0);

	if(NULL != hPopupMenu)
	{
		POINT pt;

		FillMenu(hPopupMenu);

		GetCursorPos(&pt);

		SetForegroundWindow(hwnd);

		TrackPopupMenu(hPopupMenu, TPM_LEFTALIGN | TPM_RIGHTBUTTON,
					   pt.x, pt.y,  0, hwnd, NULL);

	}

	DestroyMenu(hMenu);
}





//
// Main window OnCommand handler.
//

static
BOOL
MainWindow_OnCommand(
    IN HWND hwnd,
    IN int id,
    IN HWND hwndCtl,
    IN UINT codeNotify
   )
{
	UNREFERENCED_PARAMETER(codeNotify);
	UNREFERENCED_PARAMETER(hwndCtl);


	if((IDM_FIRSTMODE <= id) && ((IDM_FIRSTMODE + MAX_MODES) > id))
	{
		SwitchToIndex(NULL, id - IDM_FIRSTMODE);
		AdjustWS();
		return TRUE;
	}

	switch (id)
	{
	case IDM_OPTIONS:
		ChangeOptions();
		break;

	case IDM_EXIT:
		DestroyWindow(hwnd);
		break;

	case IDM_ABOUT:
		About();
		break;

	default:
		return FALSE;
	}

	return TRUE;
}



//
// Main window OnInitmenu handler.
//

static
BOOL
MainWindow_OnInitMenu(
    IN HWND hwnd,
    IN HMENU hMenu
   )
{
	UNREFERENCED_PARAMETER(hMenu);
	GetSystemMenu(hwnd, TRUE);
	ReplaceSystemMenu(GetResourceInstance(), hwnd, IDR_POPMENU, 0);
	return TRUE;
}



//
//	main window procedure
//

static
LRESULT
CALLBACK
WndProc(
    IN HWND  hwnd,
    IN UINT  uMsg,
    IN WPARAM wParam,
    IN LPARAM lParam
   )
{
	if (WM_TASKBAR_CREATED == uMsg) {

		TrayMessage(hwnd, NIM_ADD, ID_INDICATOR, NULL, NULL);
		UpdateTrayState(hwnd);
		return 0;
	}


	switch (uMsg)
	{
	case WM_CREATE:
		if(AllowedToUseTray())
			TrayMessage(hwnd, NIM_ADD, ID_INDICATOR, NULL, NULL);

		UpdateTrayState(hwnd);
		break;

	case WM_DESTROY:
		if(AllowedToUseTray())
			TrayMessage(hwnd, NIM_DELETE, ID_INDICATOR, NULL, NULL);

		PostQuitMessage(0);
		return 0;

	case WM_COMMAND:
		return HANDLE_WM_COMMAND(hwnd, wParam, lParam, MainWindow_OnCommand);

	case WM_INITMENU:
		return HANDLE_WM_INITMENU(hwnd, wParam, lParam, MainWindow_OnInitMenu);

	case WM_SYSCOMMAND:
		if(MainWindow_OnCommand(hwnd, wParam, NULL, 0))
			return 0;
		break;


	case WM_QUERYOPEN:
		return 0;

	case WM_RBUTTONUP:
		ProcessMainWindowMenu(hwnd);
		return 0;

	case _WM_NOTIFYICON:
		switch (lParam)
		{
		case WM_LBUTTONUP:
		case WM_RBUTTONUP:
			ProcessMainWindowMenu(hwnd);
			return 0;

		default:
			break;
		}

		break;

	case WM_DISPLAYCHANGE:
		UpdateTrayState(hwnd);
		break;

	}

	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}




//
// Reset mode to VGA default.
//

static
void
ResetVgaMode(void)
{
	LPDEVMODE pvm = NULL;
	UINT cModes = FillVideoModes(NULL, &pvm);

	if((cModes > 0) && (NULL != pvm))
	{
		pvm->dmBitsPerPel = 8;
		pvm->dmPelsWidth = 640;
		pvm->dmPelsHeight = 480;
		pvm->dmDisplayFrequency = 60;
		pvm->dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT | DM_DISPLAYFREQUENCY;

		ChangeDisplaySettings(pvm, CDS_UPDATEREGISTRY);

		Free(pvm);
	}

}



void
__cdecl
main(void)
{
	WNDCLASS  wc;
	int argc;

	LPWSTR* argv = CommandLineToArgvW(GetCommandLineW(), &argc);

	if((NULL != argv) && (argc > 1))
	{
		if(0 == lstrcmpi(argv[1], BASEVIDEO))
		{
			ResetVgaMode();
		}
		else
		{
			StrResMessageBox(GetThreadLocale(), GetResourceInstance(), NULL, MSG_USAGE, MSG_APPNAME, MB_ICONINFORMATION|MB_OK, argv[0]);
		}

		ExitProcess(0);
	}

	WM_TASKBAR_CREATED = RegisterWindowMessage(TB_CREATED_MESSAGE);


	FillMemory (&wc, sizeof(wc), 0);

	wc.lpfnWndProc   = WndProc;
	wc.hInstance     = GetResourceInstance();
	wc.lpszClassName = APPNAME;
	wc.hIcon = LoadIcon(GetResourceInstance(), MAKEINTRESOURCE(IDI_MAINICON));

	if(!RegisterClass(&wc))
	{
		StrResMessageBox(GetThreadLocale(), GetResourceInstance(), NULL, MSG_ERRORREGISTER, MSG_ERRORTITLE, MB_ICONSTOP|MB_OK);
	}
	else
	{
		//
		// create window
		//

		HWND hWnd;
		TCHAR szTitle[80];

		FormatString(GetThreadLocale(), GetResourceInstance(), MSG_APPNAME, szTitle, sizeof(szTitle));

		hWnd = CreateWindowEx(0, APPNAME, szTitle, WS_SYSMENU | WS_MINIMIZE | WS_OVERLAPPED,
							  CW_USEDEFAULT,CW_USEDEFAULT,0,0,
							  NULL, NULL, GetResourceInstance(), NULL);

		if(NULL == hWnd)
		{
			StrResMessageBox(GetThreadLocale(), GetResourceInstance(), NULL, MSG_ERRORCREATINGWIN, MSG_ERRORTITLE, MB_ICONSTOP|MB_OK);
		}
		else
		{
			MSG msg;

			if(!AllowedToUseTray())
			{
				ShowWindow(hWnd, SW_SHOWMINIMIZED);
			}


			//
			// Here is the place to set my working set size.
			//

			AdjustWS();

			//
			// loop
			//

			while (GetMessage(&msg, NULL, 0, 0))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}

		}
	}



	ExitProcess(0);
}


